
/*
 * Copyright 2005 Device Drivers Limited.
 * Author: kawamoto@devdrv.co.jp
 *
 * gpio.c
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#include <linux/module.h>
#include <linux/config.h>
#include <linux/vmalloc.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <asm/mach-au1x00/au1000.h>
#include <asm/mach-ekit1100/ekit1100.h>

#define BUFFER_SIZE 12
#if (PAGE_SIZE < BUF_SIZE)
#define BUFFER_SIZE PAGE_SIZE
#endif

#define GPIO_IN  0xE0000000 // GPIO[31:29]
#define GPIO_OUT 0x00F00000 // GPIO[23:20]

static char *in_data = NULL; // in_data[BUFFER_SIZE]
static char *out_data= NULL; // out_data[BUFFER_SIZE]

static BCSR * const bcsr = (BCSR *)BCSR_KSEG1_ADDR;

void gpio_init(void)
{
	outl(GPIO_OUT, SYS_OUTPUTSET);
	outl(GPIO_IN, SYS_TRIOUTCLR);
}

int gpio_in(void)
{
	return(~inl(SYS_PINSTATERD) & (GPIO_IN | GPIO_OUT));
}

void gpio_out(int d)
{
	outl(~d & GPIO_OUT, SYS_OUTPUTSET);
	outl( d & GPIO_OUT, SYS_OUTPUTCLR);
}

int gpio_read_proc(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0;

	memset(in_data, 0, BUFFER_SIZE);
	len = sprintf(in_data, "%X\n", gpio_in());
	memcpy(buf, in_data, len);

	return(len);
}

int gpio_write_proc(struct file *file, const char *buf, unsigned long count, void *data)
{
	int len;
	int d;

	len = count > BUFFER_SIZE ? BUFFER_SIZE : count;
	copy_from_user(out_data, buf, len);
	d = simple_strtoul(out_data, NULL, 16);
	gpio_out(d);
	
	return(len);
}

static void gpio_create_proc(void)
{
	struct proc_dir_entry *entry;
	entry = create_proc_entry("gpio", 0, 0);
	entry->read_proc = gpio_read_proc;
	entry->write_proc = gpio_write_proc;
}

static void gpio_remove_proc(void)
{
	remove_proc_entry("gpio", NULL);
}

int __init gpio_init_module(void)
{
	in_data = (char *)vmalloc(BUFFER_SIZE);
	if(in_data == NULL)
	{
		printk(KERN_ERR "GPIO in_data, cannot vmalloc = %d\n", BUFFER_SIZE);
		return(0);
	}
	out_data = (char *)vmalloc(BUFFER_SIZE);
	if(out_data == NULL)
	{
		printk(KERN_ERR "GPIO out_data, cannot vmalloc = %d\n", BUFFER_SIZE);
		return(0);
	}
	gpio_init();
	gpio_create_proc();
	printk("<1>GPIO init\n");
	return(0);
}	

void __exit gpio_cleanup_module(void)
{
	printk("<1>GPIO cleanup\n");
	if(in_data != NULL)
	{
		vfree(in_data);
	}
	if(out_data != NULL)
	{
		vfree(out_data);
	}
	gpio_init();
	gpio_remove_proc();
}

MODULE_DESCRIPTION("GPIO Module");
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Device Drivers Ltd.");

module_init(gpio_init_module);
module_exit(gpio_cleanup_module);
